//# ParmTableAIPS.cc: Object to hold parameters in an AIPS++ table.
//#
//# Copyright (C) 2002
//# ASTRON (Netherlands Foundation for Research in Astronomy)
//# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
//#
//# This program is free software; you can redistribute it and/or modify
//# it under the terms of the GNU General Public License as published by
//# the Free Software Foundation; either version 2 of the License, or
//# (at your option) any later version.
//#
//# This program is distributed in the hope that it will be useful,
//# but WITHOUT ANY WARRANTY; without even the implied warranty of
//# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//# GNU General Public License for more details.
//#
//# You should have received a copy of the GNU General Public License
//# along with this program; if not, write to the Free Software
//# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//#
//# $Id$

#include <lofar_config.h>
#include <BBS3/MNS/ParmTableAIPS.h>
#include <BBS3/MNS/MeqDomain.h>
#include <Common/LofarLogger.h>
#include <tables/Tables/TableDesc.h>
#include <tables/Tables/SetupNewTab.h>
#include <tables/Tables/ExprNode.h>
#include <tables/Tables/ExprNodeSet.h>
#include <tables/Tables/ScalarColumn.h>
#include <tables/Tables/ArrayColumn.h>
#include <tables/Tables/ScaColDesc.h>
#include <tables/Tables/ArrColDesc.h>
#include <tables/Tables/TableRecord.h>
#include <casa/Arrays/Matrix.h>
#include <casa/Arrays/Vector.h>
#include <casa/Arrays/ArrayUtil.h>
#include <casa/Utilities/Regex.h>
#include <casa/BasicMath/Math.h>

using namespace casa;

namespace LOFAR {

ParmTableAIPS::ParmTableAIPS (const string& tableName) :
  itsTable         (tableName),
  itsIndex         (itsTable, "NAME"),
  itsIndexName     (itsIndex.accessKey(), "NAME"),
  itsInitIndex     (0)
{
  // These things should be done in the connect
  // (just like the initialization of itsTable etc.).
  if (itsTable.keywordSet().isDefined ("DEFAULTVALUES")) {
    itsInitTable = itsTable.keywordSet().asTable ("DEFAULTVALUES");
    itsInitIndex = new ColumnsIndex (itsInitTable, "NAME");
    itsInitIndexName = RecordFieldPtr<String> (itsInitIndex->accessKey(),
					       "NAME");
  }
}

ParmTableAIPS::~ParmTableAIPS()
{
  delete itsInitIndex;
}

void ParmTableAIPS::connect()
{
  // connect was done in the constructor
}

void ParmTableAIPS::createTable (const string& tableName)
{
  TableDesc td("PSS parameter table", TableDesc::New);
  td.comment() = String("Table containing parameters for PSS");
  td.addColumn (ScalarColumnDesc<String>("NAME"));
  td.addColumn (ScalarColumnDesc<double>("STARTTIME"));
  td.addColumn (ScalarColumnDesc<double>("ENDTIME"));
  td.addColumn (ScalarColumnDesc<double>("STARTFREQ"));
  td.addColumn (ScalarColumnDesc<double>("ENDFREQ"));
  td.addColumn (ArrayColumnDesc <double>("VALUES"));
  td.addColumn (ArrayColumnDesc <double>("SIM_VALUES"));
  td.addColumn (ArrayColumnDesc <double>("SIM_PERT"));
  td.addColumn (ScalarColumnDesc<double>("TIME0"));
  td.addColumn (ScalarColumnDesc<double>("FREQ0"));
  td.addColumn (ScalarColumnDesc<double>("TIMESCALE"));
  td.addColumn (ScalarColumnDesc<double>("FREQSCALE"));
  td.addColumn (ArrayColumnDesc <bool>  ("SOLVABLE"));
  td.addColumn (ScalarColumnDesc<double>("DIFF"));
  td.addColumn (ScalarColumnDesc<bool>  ("DIFF_REL"));

  SetupNewTable newtab(tableName, td, Table::New);

  TableDesc tddef("PSS default parameter values", TableDesc::New);
  tddef.comment() = String("Table containing default parameters for PSS");
  tddef.addColumn (ScalarColumnDesc<String>("NAME"));
  tddef.addColumn (ArrayColumnDesc <double>("VALUES"));
  tddef.addColumn (ArrayColumnDesc <double>("SIM_VALUES"));
  tddef.addColumn (ArrayColumnDesc <double>("SIM_PERT"));
  tddef.addColumn (ScalarColumnDesc<double>("TIME0"));
  tddef.addColumn (ScalarColumnDesc<double>("FREQ0"));
  tddef.addColumn (ScalarColumnDesc<double>("TIMESCALE"));
  tddef.addColumn (ScalarColumnDesc<double>("FREQSCALE"));
  tddef.addColumn (ArrayColumnDesc <bool>  ("SOLVABLE"));
  tddef.addColumn (ScalarColumnDesc<double>("DIFF"));
  tddef.addColumn (ScalarColumnDesc<bool>  ("DIFF_REL"));

  SetupNewTable newdeftab(tableName+string("/DEFAULTVALUES"), tddef, Table::New);

  Table tab(newtab);
  Table deftab(newdeftab);
  tab.rwKeywordSet().defineTable("DEFAULTVALUES", deftab);  

  tab.tableInfo().setType ("MEP");
  tab.tableInfo().readmeAddLine ("PSS ME Parameter values");
  deftab.tableInfo().setType ("MEPinit");
  deftab.tableInfo().readmeAddLine ("Initial PSS ME Parameter values");


  // some descriptions could be added to the tables
}

void ParmTableAIPS::clearTable()
{
  Vector<uInt> initRows = itsInitTable.rowNumbers();
  itsInitTable.removeRow(initRows);
  Vector<uInt> rows = itsTable.rowNumbers();
  itsTable.removeRow(rows);
}

vector<MeqPolc> ParmTableAIPS::getPolcs (const string& parmName,
					 const MeqDomain& domain)
{
  vector<MeqPolc> result;
  Table sel = find (parmName, domain);
  if (sel.nrow() > 0) {
    // Get rownrs in parm table itself.
    Vector<uInt> rownrs = sel.rowNumbers (itsTable);
    ROScalarColumn<double> stCol (sel, "STARTTIME");
    ROScalarColumn<double> etCol (sel, "ENDTIME");
    ROScalarColumn<double> sfCol (sel, "STARTFREQ");
    ROScalarColumn<double> efCol (sel, "ENDFREQ");
    ROArrayColumn<bool> maskCol (sel, "SOLVABLE");
    ROArrayColumn<double> valCol (sel, "VALUES");
    ROArrayColumn<double> simvalCol (sel, "SIM_VALUES");
    ROArrayColumn<double> simpertCol (sel, "SIM_PERT");
    ROScalarColumn<double> t0Col (sel, "TIME0");
    ROScalarColumn<double> f0Col (sel, "FREQ0");
    ROScalarColumn<double> tscCol (sel, "TIMESCALE");
    ROScalarColumn<double> fscCol (sel, "FREQSCALE");
    ROScalarColumn<double> diffCol (sel, "DIFF");
    ROScalarColumn<bool> drelCol (sel, "DIFF_REL");
    for (unsigned int i=0; i<sel.nrow(); i++) {
      MeqPolc polc;
      // Set the row number as the ID, so it can be used in the put.
      polc.setID (rownrs[i]);
      if (maskCol.isDefined(i)) {
	polc.setCoeff (Matrix<double>(valCol(i)), maskCol(i));
      } else {
	polc.setCoeff (Matrix<double>(valCol(i)));
      }
      polc.setX0 (f0Col(i));
      polc.setY0 (t0Col(i));
      polc.setXScale (fscCol(i));
      polc.setYScale (tscCol(i));
      polc.setSimCoeff (Matrix<double>(simvalCol(i)));
      polc.setPertSimCoeff (Matrix<double>(simpertCol(i)));
      polc.setDomain (MeqDomain(sfCol(i), efCol(i), stCol(i), etCol(i)));
      polc.setPerturbation (diffCol(i), drelCol(i));
      result.push_back (polc);
    }
  }
  return result;
}

MeqPolc ParmTableAIPS::getInitCoeff (const string& parmName)
{
  // Try to find the default initial values in the InitialValues subtable.
  // The parameter name consists of parts (separated by dots), so the
  // parameters are categorised in that way.
  // An initial value can be defined for the full name or for a higher
  // category.
  // So look up until found or until no more parts are left.
  MeqPolc result;
  if (itsInitIndex) {
    string name = parmName;
    while (true) {
      *itsInitIndexName = name;
      Vector<uInt> rownrs = itsInitIndex->getRowNumbers();
      if (rownrs.nelements() > 0) {
	ASSERTSTR (rownrs.nelements() == 1,
		   "Too many default coefficients in parmtableAIPS");
	int row = rownrs(0);
	ROArrayColumn<bool> maskCol (itsInitTable, "SOLVABLE");
	ROArrayColumn<Double> valCol (itsInitTable, "VALUES");
	ROScalarColumn<double> t0Col (itsInitTable, "TIME0");
	ROScalarColumn<double> f0Col (itsInitTable, "FREQ0");
	ROScalarColumn<double> tscCol (itsInitTable, "TIMESCALE");
	ROScalarColumn<double> fscCol (itsInitTable, "FREQSCALE");
	ROScalarColumn<double> diffCol (itsInitTable, "DIFF");
	ROScalarColumn<bool> drelCol (itsInitTable, "DIFF_REL");
	ROArrayColumn<double> simvalCol (itsInitTable, "SIM_VALUES");
	ROArrayColumn<double> simpertCol (itsInitTable, "SIM_PERT");
	if (maskCol.isDefined(row)) {
	  result.setCoeff (Matrix<double>(valCol(row)), maskCol(row));
	} else {
	  result.setCoeff (Matrix<double>(valCol(row)));
	}
	result.setX0 (f0Col(row));
	result.setY0 (t0Col(row));
	result.setXScale (fscCol(row));
	result.setYScale (tscCol(row));
	result.setPerturbation (diffCol(row), drelCol(row));
	result.setSimCoeff (Matrix<double>(simvalCol(row)));
	result.setPertSimCoeff (Matrix<double>(simpertCol(row)));
	break;
      }
      string::size_type idx = name.rfind ('.');
      // Exit loop if no more name parts.
      if (idx == string::npos) {
	break;
      }
      // Remove last part and try again.
      name = name.substr (0, idx);
    }
  }
  return result;
}
				    
void ParmTableAIPS::putCoeff (const string& parmName,
			      MeqPolc& polc)
{
  itsTable.reopenRW();
  int rownr = polc.getID();
  if (rownr < 0) {
    if (rownr == -2) {
      // It is certainly a new row.
      putNewCoeff (parmName, polc);
    } else {
      // It might be a new row.
      putCoeffCheck (parmName, polc);
    }
  } else {
    ArrayColumn<double> valCol (itsTable, "VALUES");
    valCol.put (rownr, polc.getCoeff().getDoubleMatrix());
  }
}

void ParmTableAIPS::putCoeffCheck (const string& parmName,
				   MeqPolc& polc)
{
  const MeqDomain& domain = polc.domain();
  Table sel = find (parmName, domain);
  if (sel.nrow() > 0) {
    ASSERTSTR (sel.nrow()==1, "Parameter " << parmName <<
		 " has multiple entries for time "
		 << domain.startX() << ':' << domain.endX() << " and freq "
		 << domain.startY() << ':' << domain.endY());
    ROScalarColumn<double> stCol (sel, "STARTTIME");
    ROScalarColumn<double> etCol (sel, "ENDTIME");
    ROScalarColumn<double> sfCol (sel, "STARTFREQ");
    ROScalarColumn<double> efCol (sel, "ENDFREQ");
    ASSERTSTR (near(domain.startX(), stCol(0))  &&
	       near(domain.endX(), etCol(0))  &&
	       near(domain.startY(), sfCol(0))  &&
	       near(domain.endY(), efCol(0)),
	       "Parameter " << parmName <<
	       " has a partially instead of fully matching entry for time "
		 << domain.startX() << ':' << domain.endX() << " and freq "
		 << domain.startY() << ':' << domain.endY());
    ArrayColumn<double> valCol (sel, "VALUES");
    ArrayColumn<double> simvalCol (sel, "SIM_VALUES");
    ArrayColumn<double> simpertCol (sel, "SIM_PERT");
    valCol.put (0, polc.getCoeff().getDoubleMatrix());
  } else {
    putNewCoeff (parmName, polc);
  }
}

void ParmTableAIPS::putDefCoeff (const string& parmName,
				 MeqPolc& polc)
{
  itsInitTable.reopenRW();
  // First see if the parameter name exists at all.
  *itsInitIndexName = parmName;
  Vector<uInt> rownrs = itsInitIndex->getRowNumbers();
  if (rownrs.nelements() == 1) {
    Table sel = itsInitTable(rownrs);
    uInt rownr=0;
    ScalarColumn<String> namCol (sel, "NAME");
    ScalarColumn<double> t0Col (sel, "TIME0");
    ScalarColumn<double> f0Col (sel, "FREQ0");
    ScalarColumn<double> tscCol (sel, "TIMESCALE");
    ScalarColumn<double> fscCol (sel, "FREQSCALE");
    ScalarColumn<double> diffCol (sel, "DIFF");
    ScalarColumn<bool> drelCol (sel, "DIFF_REL");
    namCol.put (rownr, parmName);
    ArrayColumn<double> valCol (sel, "VALUES");
    ArrayColumn<double> simvalCol (sel, "SIM_VALUES");
    ArrayColumn<double> simpertCol (sel, "SIM_PERT");
    valCol.put (rownr, polc.getCoeff().getDoubleMatrix());
    simvalCol.put (rownr, polc.getSimCoeff().getDoubleMatrix());
    simpertCol.put (rownr, polc.getPertSimCoeff().getDoubleMatrix());
    f0Col.put   (rownr, polc.getX0());
    t0Col.put   (rownr, polc.getY0());
    fscCol.put   (rownr, polc.getXScale());
    tscCol.put   (rownr, polc.getYScale());
    diffCol.put (rownr, polc.getPerturbation());
    drelCol.put (rownr, polc.isRelativePerturbation());
  } else if (rownrs.nelements() == 0) {
    putNewDefCoeff(parmName, polc);
  } else {
    ASSERTSTR (false, "Too many default parms with the same name/domain")
  }
}

void ParmTableAIPS::putNewCoeff (const string& parmName, 
				 MeqPolc& polc)
{
  itsTable.reopenRW();
  uInt rownr = itsTable.nrow();
  itsTable.addRow();
  ScalarColumn<String> namCol (itsTable, "NAME");
  ScalarColumn<double> stCol (itsTable, "STARTTIME");
  ScalarColumn<double> etCol (itsTable, "ENDTIME");
  ScalarColumn<double> sfCol (itsTable, "STARTFREQ");
  ScalarColumn<double> efCol (itsTable, "ENDFREQ");
  ScalarColumn<double> t0Col (itsTable, "TIME0");
  ScalarColumn<double> f0Col (itsTable, "FREQ0");
  ScalarColumn<double> tscCol (itsTable, "TIMESCALE");
  ScalarColumn<double> fscCol (itsTable, "FREQSCALE");
  ScalarColumn<double> diffCol (itsTable, "DIFF");
  ScalarColumn<bool> drelCol (itsTable, "DIFF_REL");
  namCol.put (rownr, parmName);
  sfCol.put (rownr, polc.domain().startX());
  efCol.put (rownr, polc.domain().endX());
  stCol.put (rownr, polc.domain().startY());
  etCol.put (rownr, polc.domain().endY());
  ArrayColumn<double> valCol (itsTable, "VALUES");
  ArrayColumn<double> simvalCol (itsTable, "SIM_VALUES");
  ArrayColumn<double> simpertCol (itsTable, "SIM_PERT");
  valCol.put (rownr, polc.getCoeff().getDoubleMatrix());
  simvalCol.put (rownr, polc.getSimCoeff().getDoubleMatrix());
  simpertCol.put (rownr, polc.getPertSimCoeff().getDoubleMatrix());
  f0Col.put   (rownr, polc.getX0());
  t0Col.put   (rownr, polc.getY0());
  fscCol.put   (rownr, polc.getXScale());
  tscCol.put   (rownr, polc.getYScale());
  diffCol.put (rownr, polc.getPerturbation());
  drelCol.put (rownr, polc.isRelativePerturbation());
  // Polc is now stored here.
  polc.setID (rownr);
}


void ParmTableAIPS::putNewDefCoeff (const string& parmName, 
				    MeqPolc& polc)
{
  itsTable.reopenRW();
  itsInitTable.reopenRW();

  uInt rownr = itsInitTable.nrow();
  itsInitTable.addRow();
  ScalarColumn<String> namCol (itsInitTable, "NAME");
  ScalarColumn<double> t0Col (itsInitTable, "TIME0");
  ScalarColumn<double> f0Col (itsInitTable, "FREQ0");
  ScalarColumn<double> tscCol (itsInitTable, "TIMESCALE");
  ScalarColumn<double> fscCol (itsInitTable, "FREQSCALE");
  ScalarColumn<double> diffCol (itsInitTable, "DIFF");
  ScalarColumn<bool> drelCol (itsInitTable, "DIFF_REL");
  namCol.put (rownr, parmName);
  ArrayColumn<double> valCol (itsInitTable, "VALUES");
  ArrayColumn<double> simvalCol (itsInitTable, "SIM_VALUES");
  ArrayColumn<double> simpertCol (itsInitTable, "SIM_PERT");
  valCol.put (rownr, polc.getCoeff().getDoubleMatrix());
  simvalCol.put (rownr, polc.getSimCoeff().getDoubleMatrix());
  simpertCol.put (rownr, polc.getPertSimCoeff().getDoubleMatrix());
  f0Col.put   (rownr, polc.getX0());
  t0Col.put   (rownr, polc.getY0());
  fscCol.put   (rownr, polc.getXScale());
  tscCol.put   (rownr, polc.getYScale());
  diffCol.put (rownr, polc.getPerturbation());
  drelCol.put (rownr, polc.isRelativePerturbation());
}

Table ParmTableAIPS::find (const string& parmName,
			   const MeqDomain& domain)
{
  // First see if the parameter name exists at all.
  Table result;
  *itsIndexName = parmName;
  Vector<uInt> rownrs = itsIndex.getRowNumbers();
  if (rownrs.nelements() > 0) {
    Table sel = itsTable(rownrs);
    // Find all rows overlapping the requested domain.
    Table sel3 = sel(domain.startY() < sel.col("ENDTIME")   &&
		     domain.endY()   > sel.col("STARTTIME") &&
		     domain.startX() < sel.col("ENDFREQ")   &&
		     domain.endX()   > sel.col("STARTFREQ"));
    result = sel3;
  }
  return result;
}

vector<string> ParmTableAIPS::getSources()
{
  // Get all parm rows containing RA in the name.
  // Use the DEFAULTTABLE only if available.
  int st = 0;
  Table tab = itsInitTable;
  if (tab.isNull()) {
    tab = itsTable;
    st = 1;
  }
  vector<string> nams;
  for (int i=st; i<2; i++) {
    TableExprNode expr(tab.col("NAME") == Regex(Regex::fromPattern("RA.*")));
    Table sel = tab(expr);
    if (sel.nrow() > 0) {
      // Sort them uniquely on sourcenr.
      Table sor = sel.sort("NAME", Sort::Ascending,
			   Sort::QuickSort | Sort::NoDuplicates);
      ROScalarColumn<String> namCol(sor, "NAME");
      Vector<String> names = namCol.getColumn();
      nams.reserve (nams.size() + sor.nrow());
      for (unsigned int i=0; i<names.nelements(); i++) {
	nams.push_back (names(i));
      }
    }
    tab = itsTable;
  }
  LOG_TRACE_STAT_STR("Finished retreiving "<<nams.size()<<" sources");
  return nams;
}

void ParmTableAIPS::unlock()
{
  itsTable.unlock();
}

}
