//#  WH_Dump.cc: Output workholder for BG Correlator application
//#
//#  Copyright (C) 2002-2004
//#  ASTRON (Netherlands Foundation for Research in Astronomy)
//#  P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, swe@astron.nl
//#
//#
//#  $Id$

#include <lofar_config.h>

#include <stdio.h>           // for sprintf
#include <sys/types.h>       // for file output
#include <fcntl.h>           // for file output
#include <math.h>            // for sqrt()

// General includes
#include <Common/LofarLogger.h>
#include <Common/KeyValueMap.h>

// Application specific includes
#include <WH_Dump.h>
#include <DH_Vis.h>

using namespace LOFAR;


WH_Dump::WH_Dump(const string& name, const KeyValueMap& kvm, const string& fileName)
  : WorkHolder(kvm.getInt("NoWH_Correlator",7)/kvm.getInt("NoWH_Dump", 2), 0, name, "WH_Dump"),
    itsOutputFile(0),
    itsBandwidth(0),
    itsKvm(kvm)
{
  itsOutputFileName = fileName;
  itsNelements      = kvm.getInt("NoWH_RSP", 2);
  itsNchannels      = kvm.getInt("NoRSPBeamlets", 92)/kvm.getInt("NoWH_Correlator", 92);
  itsNpolarisations = kvm.getInt("polarisations", 2);
  
  char DHName[20];
  for (int i=0; i<itsNinputs; i++) {
    sprintf(DHName, "input_%3d_of_%3d", i, itsNinputs);
    getDataManager().addInDataHolder(i, new DH_Vis(DHName, 
						   itsNelements, 
						   itsNchannels, 
						   itsNpolarisations));
  }

  itsLastTime.tv_sec = 0;
  itsLastTime.tv_usec = 0;
  // From now on the number of polarisations is squared because we
  // are behind the correlator. The DataHolder that is added above
  // does the multiplication in its own constructor
  itsNpolarisations *= itsNpolarisations;
}

WH_Dump::~WH_Dump() {
}

WorkHolder* WH_Dump::construct (const string& name, 
				const KeyValueMap& kvm,
				const string& fileName)
{
  return new WH_Dump(name, kvm, fileName);
}

WH_Dump* WH_Dump::make(const string& name) {
  return new WH_Dump(name, itsKvm, itsOutputFileName); 
}

void WH_Dump::preprocess() {
  int openFlags = O_WRONLY | O_CREAT | O_TRUNC | O_LARGEFILE;
  mode_t permissions = S_IREAD | S_IWRITE | S_IRGRP | S_IROTH;
  itsOutputFile = open(itsOutputFileName.c_str(), openFlags, permissions);
  ASSERTSTR(itsOutputFile != -1, "Could not open outputfile " << itsOutputFileName);
}

void WH_Dump::process() {

  struct timeval newTime;
  long recSize = 0;
  long written = 0;
  long totalWritten = 0;
  long totalWrittenKB = 0;
  DH_Vis* dhp;

  // Right now we write to disk per frequency channel
  // The file consists of a multidimensional array: channels*station*station*npolariations
  // Every block of stations*stations holds the real values in the upper triangle and the 
  // imaginary values in the lower triangle. 

  // This will probably give some performance problems. If it does we can do the following improvements:
  // 1 Put the real and imaginairy values together and make the dimensions of the array dynamic. This way no
  //   extra memory is used and we can read the complex numbers as a whole.
  // 2 Make the format at the output of the correlator the same as in the file or vice versa. Then we can read
  //   and write the data in very large blocks.

  // We need some way to store metadata. Maybe in another file (or a piece of paper :)

  DH_Vis::BufferPrimitive freqBlock[itsNelements*itsNelements*itsNpolarisations];
  
  for (int i=0; i<itsNinputs; i++) {
    dhp = (DH_Vis*)getDataManager().getInHolder(i);
    recSize += dhp->getBufSize()*sizeof(DH_Vis::BufferType);
    for (int channel=0; channel<itsNchannels; channel++) {
#if 1
      int noAns = itsNelements * std::sqrt((double)itsNpolarisations);
      memset(freqBlock, 0, itsNelements*itsNelements*itsNpolarisations * sizeof(DH_Vis::BufferPrimitive));
      for (int an1=0; an1 < noAns; an1++) {
 	for (int an2=0; an2 < an1; an2++) {
 	  freqBlock[ an1 * noAns + an2 ] = dhp->getBufferElement(an1/2, an2/1, i, 2 * (an1%2) +  an2%2)->imag();
 	}
 	for (int an2=an1; an2 < noAns; an2++) {
 	  freqBlock[ an1 * noAns + an2 ] = dhp->getBufferElement(an2/2, an1/2, i, 2 * (an2%2) + an1%2)->real();
 	}
      }
#else
      for (int el1=0; el1<itsNelements; el1++){
	for (int el2=0; el2<=el1; el2++){
	  int offsetIm = el1 * (itsNelements * itsNpolarisations) + el2 * itsNpolarisations;
	  int offsetRe = el2 * (itsNelements * itsNpolarisations) + el1 * itsNpolarisations;
	  for (int pol=0; pol<itsNpolarisations; pol++){
	    freqBlock[ offsetIm + pol ] = dhp->getBufferElement(el1, el2, channel, pol)->imag();
	    freqBlock[ offsetRe + pol ] = dhp->getBufferElement(el1, el2, channel, pol)->real();
	  }
	}
      }
#endif
      written = write(itsOutputFile, freqBlock, itsNelements * itsNelements * itsNpolarisations * sizeof(DH_Vis::BufferPrimitive));
      
      if (written == -1) {
	cerr<<"Something went wrong during write!"<<endl;
      }
      totalWritten += written;
    }
  }
  recSize = recSize / 1024; // from now recSize is in kB
  totalWrittenKB = totalWritten / 1024;
  gettimeofday(&newTime, NULL);
  if (itsLastTime.tv_sec != 0) {
    double elapsed = newTime.tv_sec - itsLastTime.tv_sec + 10e-6*(newTime.tv_usec - itsLastTime.tv_usec);
    itsBandwidth = recSize/elapsed;
    cout << "Received " 
	 << recSize << " kB (" 
	 << itsBandwidth / elapsed << " kB/s), written " 
	 << totalWrittenKB << " kB (" 
	 << totalWrittenKB / elapsed <<" kB/s)"
	 << endl;    
  }

  memcpy(&itsLastTime, &newTime, sizeof(struct timeval));
}

void WH_Dump::postProcess() {
  close(itsOutputFile);
  itsOutputFile = 0;
}

void WH_Dump::dump() const{
  cout<<"DUMP OF WH_DUMP: "<<getName()<<endl;
  DH_Vis* dhp;
  for (int i=0; i<itsNinputs; i++) {
    dhp = (DH_Vis*)getDataManager().getInHolder(i);
    for (int channel=0; channel<itsNchannels; channel++) {
      cout<<"channel: "<< channel <<" of input: "<<i<<endl;
      for (int el1=0; el1<itsNelements; el1++){
	cout<<el1<<": ";
	for (int el2=0; el2<itsNelements; el2++){
	  int offset = el2 * (itsNelements * itsNpolarisations) + el1 * itsNpolarisations;
	  for (int pol=0; pol<itsNpolarisations; pol++){
	    cout<<*dhp->getBufferElement(el1, el2, channel, pol)<<" ";
	  }
	  cout<<"  ";
	}
	cout<<endl;
      }
    }
  }
}
		 
